# example: A ring-loaded Mach-Zehnder interferometer.
# A directiopnal coupler is used for splitter and combiner
# and each MZI arm has a ring in it.

from ipkiss3 import all as i3

from ring import RingResonator
from dircoup import BentDirectionalCoupler

TECH = i3.TECH


class RingLoadedMZI(i3.PCell):
    """A Mach-Zehnder with a ring resonator in each arm"""

    _name_prefix = "MZI"
    # 1. Define the rings and splitters as childcells.
    # We add a type restriction to ensure the passed object is a RingResonator/BentDirectionalCoupler.

    ring1 = i3.ChildCellProperty(restriction=i3.RestrictType(RingResonator))
    ring2 = i3.ChildCellProperty(restriction=i3.RestrictType(RingResonator))
    splitter = i3.ChildCellProperty(restriction=i3.RestrictType(BentDirectionalCoupler))
    combiner = i3.ChildCellProperty(restriction=i3.RestrictType(BentDirectionalCoupler))

    # 2. We define a waveguide template and the waveguide cells
    wg_template = i3.WaveguideTemplateProperty(default=i3.TECH.PCELLS.WG.DEFAULT)
    wgs = i3.ChildCellListProperty(doc="list of waveguides")

    # 3. We first define the 4 waveguides that we need at the PCELL level.
    def _default_wgs(self):
        w1 = i3.Waveguide(name=self.name + "_w1", trace_template=self.wg_template)
        w2 = i3.Waveguide(name=self.name + "_w2", trace_template=self.wg_template)
        w3 = i3.Waveguide(name=self.name + "_w3", trace_template=self.wg_template)
        w4 = i3.Waveguide(name=self.name + "_w4", trace_template=self.wg_template)
        return w1, w2, w3, w4

    class Layout(i3.LayoutView):
        bend_radius = i3.PositiveNumberProperty(default=i3.TECH.WG.BEND_RADIUS)

        # 4. Placing the components
        @i3.cache()
        def _get_components(self):
            si_splitter = self.splitter.size_info()  # object with contours of the component
            si_combiner = self.combiner.size_info()  # object with contours of the component
            si_ring1 = self.ring1.size_info()  # object with contours of the component
            si_ring2 = self.ring2.size_info()  # object with contours of the component

            # spacing between components to allow easy routing (can be optimized)
            spacing = 2 * self.bend_radius + 3 * TECH.WG.SHORT_STRAIGHT

            # splitter
            t_splitter = i3.IdentityTransform()  # place the splitter in (0,0)
            splitter = i3.SRef(reference=self.splitter, name="splitter", transformation=t_splitter)

            # ring1
            t_ring1 = i3.Translation(
                (si_splitter.east - si_ring1.west + spacing, si_splitter.north - si_ring1.south + spacing)
            )
            ring1 = i3.SRef(reference=self.ring1, name="ring1", transformation=t_ring1)

            # ring2
            t_ring2 = i3.VMirror() + i3.Translation(
                (si_splitter.east - si_ring2.west + spacing, si_splitter.south + si_ring2.south - spacing)
            )
            ring2 = i3.SRef(reference=self.ring2, name="ring2", transformation=t_ring2)

            # combiner
            si_rings = si_ring1.transform(t_ring1) + si_ring2.transform(t_ring2)
            t_combiner = i3.Translation((si_rings.east - si_combiner.west + spacing, 0.0))
            combiner = i3.SRef(reference=self.combiner, name="combiner", transformation=t_combiner)

            return splitter, ring1, ring2, combiner

        # 5. Getting the layout view of the waveguides.
        def _default_wgs(self):
            splitter, ring1, ring2, combiner = self._get_components()
            w1_cell, w2_cell, w3_cell, w4_cell = self.cell.wgs
            wg_template = self.wg_template

            # Defining the north arm
            w1_layout = w1_cell.get_default_view(i3.LayoutView)

            # We set the shape using RouteManhattan and the ports.
            w1_layout.set(
                shape=i3.RouteManhattan(
                    start_port=splitter.ports["out_1"], end_port=ring1.ports["in"], bend_radius=self.bend_radius
                ),
                trace_template=wg_template,
            )

            w2_layout = w2_cell.get_default_view(i3.LayoutView)
            w2_layout.set(
                shape=i3.RouteManhattan(
                    start_port=ring1.ports["out"], end_port=combiner.ports["in_1"], bend_radius=self.bend_radius
                ),
                trace_template=wg_template,
            )

            # Defining the south arm
            w3_layout = w3_cell.get_default_view(i3.LayoutView)
            w3_layout.set(
                shape=i3.RouteManhattan(
                    start_port=splitter.ports["out_2"], end_port=ring2.ports["in"], bend_radius=self.bend_radius
                ),
                trace_template=wg_template,
            )

            w4_layout = w4_cell.get_default_view(i3.LayoutView)
            w4_layout.set(
                shape=i3.RouteManhattan(
                    start_port=ring2.ports["out"], end_port=combiner.ports["in_2"], bend_radius=self.bend_radius
                ),
                trace_template=wg_template,
            )

            return w1_layout, w2_layout, w3_layout, w4_layout

        # 6. Placing both the component and the waveguides in the layout.
        def _generate_instances(self, insts):
            insts += self._get_components()
            w1_layout, w2_layout, w3_layout, w4_layout = self.wgs

            insts += i3.SRef(reference=w1_layout, name="w1")
            insts += i3.SRef(reference=w2_layout, name="w2")
            insts += i3.SRef(reference=w3_layout, name="w3")
            insts += i3.SRef(reference=w4_layout, name="w4")
            return insts

        # 7. Generating the ports of the rlmzi
        def _generate_ports(self, prts):
            instances = self.instances
            prts += instances["splitter"].ports["in_1"]
            prts += instances["splitter"].ports["in_2"]
            prts += instances["combiner"].ports["out_1"]
            prts += instances["combiner"].ports["out_2"]
            return prts
